/* cnic.c: CNIC UIO uIP user space stack
 *
 * Copyright (c) 2004-2010 Broadcom Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * Written by: Benjamin Li  (benli@broadcom.com)
 */

#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <linux/limits.h>
#include <netinet/if_ether.h>
#include <netinet/in.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <linux/netlink.h>
#include <sys/ioctl.h>
 #include <sys/time.h>
#include <sys/types.h>
#include <sys/user.h>
#include <sys/socket.h>

#include "uip_arp.h"
#include "nic.h"
#include "nic_utils.h"
#include "logger.h"
#include "options.h"

#include "cnic.h"
#include "iscsi_if.h"

/*******************************************************************************
 * Constants
 ******************************************************************************/
#define PFX "CNIC "

static const uip_ip6addr_t all_ones_addr6 =
  {0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff};

/*******************************************************************************
 * Constants shared between the bnx2 and bnx2x modules
 ******************************************************************************/
const char bnx2i_library_transport_name[] = "bnx2i";
const size_t bnx2i_library_transport_name_size = 
	sizeof(bnx2i_library_transport_name);

/******************************************************************************
 * Netlink Functions
 ******************************************************************************/

static int cnic_arp_send(nic_t *nic, nic_interface_t *nic_iface, int fd,
			 __u8 * mac_addr, __u32 ip_addr, char *addr_str)
{
	struct ether_header *eth;
	struct ether_arp *arp;
	__u32 dst_ip = ip_addr;
	int pkt_size = sizeof(*eth) + sizeof(*arp);
	int rc;
	struct in_addr addr;
	static const uint8_t multicast_mac[] =
		{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };

	rc = pthread_mutex_trylock(&nic->xmit_mutex);
	if (rc != 0) {
		LOG_DEBUG(PFX "%s: could not get xmit_mutex", nic->log_name);
		return -EAGAIN;
	}

	eth = (*nic->ops->get_tx_pkt)(nic);
	if (eth == NULL) {
		LOG_WARN(PFX "%s: couldn't get tx packet", nic->log_name);
		return -EAGAIN;
	}

	nic_fill_ethernet_header(nic_iface, eth,
				 nic->mac_addr, multicast_mac,
				 &pkt_size, &arp, ETHERTYPE_ARP);

	arp->arp_hrd = htons(ARPHRD_ETHER);
	arp->arp_pro = htons(ETHERTYPE_IP);
	arp->arp_hln = ETH_ALEN;
	arp->arp_pln = 4;
	arp->arp_op = htons(ARPOP_REQUEST);
	memcpy(arp->arp_sha, nic->mac_addr, ETH_ALEN);

	/*  Copy the IP address's into the ARP response */
	memcpy(arp->arp_spa, nic_iface->ustack.hostaddr, 4);
	memcpy(arp->arp_tpa, &dst_ip, 4);

	(*nic->nic_library->ops->start_xmit)(nic, pkt_size);

	memcpy(&addr.s_addr, &dst_ip, sizeof(addr.s_addr));
	LOG_DEBUG(PFX "%s: Sent cnic arp request for IP: %s",
		       nic->log_name, addr_str);

	return 0;
}

static int cnic_neigh_soliciation_send(nic_t *nic,
				       nic_interface_t *nic_iface, int fd,
				       __u8 * mac_addr,
				       struct in6_addr *addr6_dst,
				       char *addr_str)
{
	struct ether_header *eth;
	struct ip6_hdr *ipv6_hdr;
	struct nd_neighbor_solicit *sol;
	struct nd_opt_hdr *opt_hdr;
	char *data;
	int rc;
	int pkt_size = sizeof(*eth) + sizeof(*ipv6_hdr) + sizeof(*sol) +
		       sizeof(*opt_hdr) + 6;

	struct in6_addr multi = 
		{ { { 0xff, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		      0x00, 0x00, 0x00, 0x01, 0xff, 0x00, 0x00, 0x00 } } };
	static u_int8_t ipv6_neigh_dhost[ETH_ALEN] =
		{ 0x33, 0x33, 0x00, 0x00, 0x00, 0x00 };

	/*  Prepare the IPv6 multicast destination address */
	multi.s6_addr[13] |= addr6_dst->s6_addr[13];
	multi.s6_addr16[7] = addr6_dst->s6_addr16[7];

	/*  Prepare the proper IPv6 Multicast Address */
	ipv6_neigh_dhost[5] = multi.s6_addr[15];
	ipv6_neigh_dhost[4] = multi.s6_addr[14];
	ipv6_neigh_dhost[3] = multi.s6_addr[13];
	ipv6_neigh_dhost[2] = multi.s6_addr[12];

	/*  Set multicast bit in MAC address */
	ipv6_neigh_dhost[2] |= 0x01;

	rc = pthread_mutex_trylock(&nic->xmit_mutex);
	if (rc != 0) {
		LOG_DEBUG(PFX "%s: could not get xmit_mutex", nic->log_name);
		return -EAGAIN;
	}

	/*  Build the ethernet header */
	eth = (*nic->ops->get_tx_pkt)(nic);
	if (eth == NULL) {
		LOG_WARN(PFX "%s: couldn't get tx packet", nic->log_name);
		return -EAGAIN;
	}

	 nic_fill_ethernet_header(nic_iface, eth,
				  nic->mac_addr, ipv6_neigh_dhost,
				  &pkt_size, &ipv6_hdr, ETHERTYPE_IPV6);

	/*  Prepare the IPv6 header */
	ipv6_hdr->ip6_vfc  = 0x6 << 4;
	ipv6_hdr->ip6_plen = htons((sizeof(*sol) + sizeof(*opt_hdr) + 6)); /*  Add optionial header */
	ipv6_hdr->ip6_hlim = 255;
	ipv6_hdr->ip6_nxt  = IPPROTO_ICMPV6;

	memcpy(ipv6_hdr->ip6_dst.s6_addr, multi.s6_addr,
	       sizeof(sol->nd_ns_target.s6_addr));

	/*  Depending on the IPv6 address of the target we will need to
	 *  determine whether we use the assigned IPv6 address or the
	 *  link local IPv6 address */
	if (is_ipv6_link_local_address(addr6_dst->s6_addr))
		memcpy(ipv6_hdr->ip6_src.s6_addr,
		       nic_iface->ustack.link_local_addr,
		       sizeof(struct in6_addr));
	else
		memcpy(ipv6_hdr->ip6_src.s6_addr, nic_iface->ustack.hostaddr6,
		       sizeof(struct in6_addr));

	/*  Prepare the ICMPv6 header */
	sol = (struct nd_neighbor_solicit *)(ipv6_hdr + 1);
	sol->nd_ns_hdr.icmp6_type = ND_NEIGHBOR_SOLICIT;
	sol->nd_ns_hdr.icmp6_code = 0;
	sol->nd_ns_hdr.icmp6_cksum = 0;
	memcpy(sol->nd_ns_target.s6_addr, addr6_dst->s6_addr,
	       sizeof(sol->nd_ns_target.s6_addr));

	/*  Prepare the ICMPv6 Option header */
	opt_hdr = (struct nd_opt_hdr *) (sol + 1);
	opt_hdr->nd_opt_type = ND_OPT_SOURCE_LINKADDR;
	opt_hdr->nd_opt_len = 8 >> 3;
	opt_hdr++;

	/*  Copy the Link Layer address */
	data = (char *) opt_hdr;
	memcpy(data, nic->mac_addr, sizeof(nic->mac_addr));

	sol->nd_ns_hdr.icmp6_cksum = ~icmpv6_checksum((uint8_t *) ipv6_hdr);

	(*nic->nic_library->ops->start_xmit)(nic, pkt_size);

	LOG_DEBUG(PFX "%s: Sent cnic ICMPv6 neighbor request %s",
		  nic->log_name, addr_str);

	return 0;
}


static int cnic_nl_neigh_rsp(nic_t *nic, int fd,
			     struct iscsi_uevent *ev,
			     struct iscsi_path *path_req,
			     __u8 *mac_addr,
			     nic_interface_t *nic_iface,
			     int status, int type)
{
	int rc;
	uint8_t *ret_buf;
	struct iscsi_uevent *ret_ev;
	struct iscsi_path *path_rsp;
	struct sockaddr_nl dest_addr;
	char addr_dst_str[INET6_ADDRSTRLEN];

        memset(&dest_addr, 0, sizeof(dest_addr));
        dest_addr.nl_family = AF_NETLINK;
        dest_addr.nl_pid = 0;
        dest_addr.nl_groups = 0; /* unicast */

	ret_buf = calloc(1, NLMSG_SPACE(sizeof(struct iscsi_uevent) + 256));
	if (ret_buf == NULL) {
                LOG_ERR(PFX "Could not allocate memory for path req resposne");
		return -ENOMEM;
	}

        memset(ret_buf, 0, NLMSG_SPACE(sizeof(struct iscsi_uevent) + 256));

	/*  prepare the iscsi_uevent buffer */
	ret_ev = (struct iscsi_uevent *)ret_buf;
	ret_ev->type = ISCSI_UEVENT_PATH_UPDATE;
	ret_ev->transport_handle = ev->transport_handle;
	ret_ev->u.set_path.host_no = ev->r.req_path.host_no;

	/*  Prepare the iscsi_path buffer */
	path_rsp = (struct iscsi_path *)(ret_buf + sizeof(*ret_ev));
	path_rsp->handle = path_req->handle;
	if (type == AF_INET) {
		path_rsp->ip_addr_len = 4;
		memcpy(&path_rsp->src.v4_addr, &nic_iface->ustack.hostaddr,
		       sizeof(nic_iface->ustack.hostaddr));

		inet_ntop(AF_INET, &path_rsp->src.v4_addr,
			  addr_dst_str, sizeof(addr_dst_str));
	} else {
		u8_t *src_ipv6;

		/*  Depending on the IPv6 address of the target we will need to
		 *  determine whether we use the assigned IPv6 address or the
		 *  link local IPv6 address */
		if (is_ipv6_link_local_address(&path_req->dst.v6_addr))
			src_ipv6 = nic_iface->ustack.link_local_addr;
		else
			src_ipv6 = nic_iface->ustack.hostaddr6;

		path_rsp->ip_addr_len = 16;
		memcpy(&path_rsp->src.v6_addr, src_ipv6,
		       sizeof(nic_iface->ustack.hostaddr6));

		inet_ntop(AF_INET6, &path_rsp->src.v6_addr,
			  addr_dst_str, sizeof(addr_dst_str));
	}
	memcpy(path_rsp->mac_addr, mac_addr, 6);
	path_rsp->vlan_id = path_req->vlan_id;
	path_rsp->pmtu    = path_req->pmtu;

        rc = __kipc_call(fd, ret_ev, sizeof(*ret_ev) + sizeof(*path_rsp));
        if (rc > 0) {
                LOG_DEBUG(PFX "neighbor reply sent back to kernel "
			      "%s at %02x:%02x:%02x:%02x:%02x:%02x",
			      addr_dst_str,
			      mac_addr[0], mac_addr[1],
			      mac_addr[2], mac_addr[3],
			      mac_addr[4], mac_addr[5]);

	} else {
                LOG_ERR(PFX "send neighbor reply failed: %d", rc);
	}

	free(ret_buf);

	return rc;
}

const static struct timeval tp_wait = {
	.tv_sec = 0,
	.tv_usec = 250000,
};

/**
 * cnic_handle_ipv4_iscsi_path_req() - This function will handle the IPv4
 * 				       path req calls the bnx2i kernel module
 * @param nic - The nic the message is directed towards
 * @param fd  - The file descriptor to be used to extract the private data
 * @param ev  - The iscsi_uevent
 * @param buf - The private message buffer
 * @param buf_len - The private message buffer length
 */
int cnic_handle_ipv4_iscsi_path_req(nic_t *nic, int fd, struct iscsi_uevent *ev,
				    struct iscsi_path *path, size_t buf_len) 
{
	nic_interface_t *nic_iface;
	struct in_addr src_addr, dst_addr, 
		       src_matching_addr, dst_matching_addr, netmask;
	__u8 mac_addr[6];
	int rc;
	uint16_t arp_retry;
	int status = 0;

	memset(mac_addr, 0, sizeof(mac_addr));

	pthread_mutex_lock(&nic_list_mutex);

	/*  Find the proper interface via VLAN id */
	nic_iface = nic_find_nic_iface_protocol(nic, path->vlan_id, AF_INET);
	if (nic_iface == NULL) {
		nic_iface = nic_find_nic_iface_protocol(nic, 0, AF_INET);
		if (nic_iface == NULL) {
			pthread_mutex_unlock(&nic_list_mutex);
			LOG_ERR(PFX "%s: Couldn't find net_iface vlan_id: %d",
				nic->log_name, path->vlan_id);
			return -EINVAL;
		}

		nic_iface->vlan_id = path->vlan_id;
	}

#define MAX_ARP_RETRY 4

	memcpy(&dst_addr, &path->dst.v4_addr, sizeof(dst_addr));
	memcpy(&src_addr, &nic_iface->ustack.hostaddr, sizeof(src_addr));

	if (nic_iface->ustack.netmask)
		memcpy(&netmask.s_addr, &nic_iface->ustack.netmask,
		       sizeof(src_addr));
	else
		netmask.s_addr = calculate_default_netmask(dst_addr.s_addr);
 
	src_matching_addr.s_addr = src_addr.s_addr & netmask.s_addr;
	dst_matching_addr.s_addr = dst_addr.s_addr & netmask.s_addr;
 
	if (src_matching_addr.s_addr != dst_matching_addr.s_addr) {
		/*  If there is an assigned gateway address then use it
		 *  if the source address doesn't match */
		if (nic_iface->ustack.default_route_addr != 0) {
			memcpy(&dst_addr,
			       &nic_iface->ustack.default_route_addr,
			       sizeof(dst_addr));
		} else {
			arp_retry = MAX_ARP_RETRY;
			goto done;
		}
	}


#define MAX_ARP_RETRY 4
	arp_retry = 0;

	rc = uip_lookup_arp_entry(dst_addr.s_addr, mac_addr);
	if (rc != 0) {
		while ((arp_retry < MAX_ARP_RETRY) && (event_loop_stop == 0)) {
			char *dst_addr_str;
			int count;
			struct timespec ts;
			struct timeval tp;
			struct timeval tp_abs;

			dst_addr_str = inet_ntoa(dst_addr);

			LOG_INFO(PFX "%s: Didn't find IPv4: '%s' in ARP table",
				 nic->log_name, dst_addr_str);
			rc = cnic_arp_send(nic, nic_iface, fd,
					   mac_addr,
					   dst_addr.s_addr, dst_addr_str);
			if (rc != 0) {
				status = -EIO;
				goto done;
			}


			for (count=0; count<8; count++) {
				/* Convert from timeval to timespec */
				rc =  gettimeofday(&tp, NULL);
			
				timeradd(&tp, &tp_wait, &tp_abs);

				ts.tv_sec  = tp_abs.tv_sec;
				ts.tv_nsec = tp_abs.tv_usec * 1000;

				pthread_mutex_unlock(&nic_list_mutex);
		                rc = pthread_cond_timedwait(&nl_process_if_down_cond,
                	                            &nl_process_mutex, &ts);

				if (rc == ETIMEDOUT) {
					pthread_mutex_unlock(&nl_process_mutex);

					if (pthread_mutex_trylock(
						&nic_list_mutex) != 0) {
						arp_retry = MAX_ARP_RETRY;
						goto done;

					}

					rc = uip_lookup_arp_entry(dst_addr.s_addr,
								  mac_addr);
					if (rc == 0)
						goto done;
				} else {
					nl_process_if_down = 0;
					pthread_mutex_unlock(&nl_process_mutex);
					
					arp_retry = MAX_ARP_RETRY;
					goto done;
					
				}
			}

			arp_retry++;
		}
	}

done:

	if (arp_retry >= MAX_ARP_RETRY) {
		status = -EIO;
		rc = -EIO;
	}

	if (status != 0 || rc != 0)
		pthread_mutex_unlock(&nic->xmit_mutex);

	cnic_nl_neigh_rsp(nic, fd, ev, path, mac_addr,
			  nic_iface, status, AF_INET);

	pthread_mutex_unlock(&nic_list_mutex);

	return rc;
}

/**
 * cnic_handle_ipv6_iscsi_path_req() - This function will handle the IPv4
 * 				       path req calls the bnx2i kernel module
 * @param nic - The nic the message is directed towards
 * @param fd  - The file descriptor to be used to extract the private data
 * @param ev  - The iscsi_uevent
 * @param buf - The private message buffer
 * @param buf_len - The private message buffer length
 */
int cnic_handle_ipv6_iscsi_path_req(nic_t *nic, int fd, struct iscsi_uevent *ev,
				    struct iscsi_path *path, size_t buf_len) 
{
	nic_interface_t *nic_iface;
	__u8 mac_addr[6];
	int rc, i;
	uint16_t neighbor_retry;
	int status = 0;
	char addr_dst_str[INET6_ADDRSTRLEN];
	struct in6_addr src_addr, dst_addr, 
		        src_matching_addr, dst_matching_addr, netmask;

	memset(mac_addr, 0, sizeof(mac_addr));

	inet_ntop(AF_INET6, &path->dst.v6_addr,
		  addr_dst_str, sizeof(addr_dst_str));

	pthread_mutex_lock(&nic_list_mutex);

	/*  Find the proper interface via VLAN id */
	nic_iface = nic_find_nic_iface_protocol(nic, path->vlan_id, AF_INET6);
	if (nic_iface == NULL) {
		nic_iface = nic_find_nic_iface_protocol(nic, 0, AF_INET6);
		if (nic_iface == NULL) {
			pthread_mutex_unlock(&nic_list_mutex);
			LOG_ERR(PFX "%s: Couldn't find net_iface vlan_id: %d",
				nic->log_name, path->vlan_id);
			return -EINVAL;
		}

		nic_iface->vlan_id = path->vlan_id;
	}

	memcpy(&dst_addr, &path->dst.v6_addr, sizeof(dst_addr));
	memcpy(&src_addr, &nic_iface->ustack.hostaddr6, sizeof(src_addr));

	if (nic_iface->ustack.netmask6)
		memcpy(&netmask.s6_addr, nic_iface->ustack.netmask6,
		       sizeof(netmask.s6_addr));
	else
		memcpy(&netmask.s6_addr, all_zeroes_addr6,
		       sizeof(netmask.s6_addr));
 
	for (i = 0; i < 4; i++) {
		src_matching_addr.s6_addr32[i] = src_addr.s6_addr32[i] &
						 netmask.s6_addr32[i];
		dst_matching_addr.s6_addr32[i] = dst_addr.s6_addr32[i] &
						 netmask.s6_addr32[i];
		if (src_matching_addr.s6_addr32[i] !=
					dst_matching_addr.s6_addr32[i]) {
			neighbor_retry = MAX_ARP_RETRY;
			goto done;
		}
 	}

	LOG_DEBUG(PFX "%s: Preparing to send IPv6 neighbor solicitation "
		      "to dst: '%s'",
		  nic->log_name, addr_dst_str);

#define MAX_ARP_RETRY 4
	neighbor_retry = 0;

	rc = uip_neighbor_lookup(&nic_iface->ustack, &path->dst.v6_addr,
				 mac_addr);
	if (rc != 0) { 
		while ((neighbor_retry < MAX_ARP_RETRY) &&
		       (event_loop_stop == 0)) {
			int count;
			struct timespec ts;
			struct timeval tp;
			struct timeval tp_abs;

			LOG_INFO(PFX "%s: Didn't find IPv6: '%s'\n",
				 nic->log_name, addr_dst_str);

			rc = cnic_neigh_soliciation_send(nic, nic_iface, fd,
							 mac_addr,
							 &path->dst.v6_addr,
							 addr_dst_str);
			if (rc != 0) {
				status = -EIO;
				goto done;
			}

			for (count=0; count<8; count++) {
				/* Convert from timeval to timespec */
				rc =  gettimeofday(&tp, NULL);
			
				timeradd(&tp, &tp_wait, &tp_abs);

				ts.tv_sec  = tp_abs.tv_sec;
				ts.tv_nsec = tp_abs.tv_usec * 1000;

				pthread_mutex_unlock(&nic_list_mutex);
		                rc = pthread_cond_timedwait(&nl_process_if_down_cond,
                	                            &nl_process_mutex, &ts);

				if (rc == ETIMEDOUT) {
					pthread_mutex_unlock(&nl_process_mutex);

					if (pthread_mutex_trylock(
						&nic_list_mutex) != 0) {
						neighbor_retry = MAX_ARP_RETRY;
						goto done;

					}

					rc = uip_neighbor_lookup(
							&nic_iface->ustack,
							&path->dst.v6_addr,
							mac_addr);

					if (rc == 0)
						goto done;
				} else {
					nl_process_if_down = 0;
					pthread_mutex_unlock(&nl_process_mutex);
					
					neighbor_retry = MAX_ARP_RETRY;
					goto done;
					
				}
			}

			neighbor_retry++;
		}
	}

done:
	if (neighbor_retry >= MAX_ARP_RETRY) {
		status = -EIO;
		rc = -EIO;
	}

	if (status != 0 || rc != 0)
		pthread_mutex_unlock(&nic->xmit_mutex);

	cnic_nl_neigh_rsp(nic, fd, ev, path, mac_addr,
			  nic_iface, status, AF_INET6);

	pthread_mutex_unlock(&nic_list_mutex);

	return rc;

	return 0;
}

/**
 * cnic_handle_iscsi_path_req() - This function will handle the path req calls 
 *                                the bnx2i kernel module
 * @param nic - The nic the message is directed towards
 * @param fd  - The file descriptor to be used to extract the private data
 * @param ev  - The iscsi_uevent
 * @param buf - The private message buffer
 * @param buf_len - The private message buffer length
 */
int cnic_handle_iscsi_path_req(nic_t *nic, int fd, struct iscsi_uevent *ev,
			       struct iscsi_path *path, size_t buf_len) 
{

	LOG_DEBUG(PFX "%s: Netlink message with VLAN ID: %d, path MTU: %d "
		      "minor: %d ip_addr_len: %d",
		   nic->log_name, path->vlan_id, path->pmtu, 0/* TODO FIX */,
		   path->ip_addr_len);

	if (path->ip_addr_len == 4)
		return cnic_handle_ipv4_iscsi_path_req(nic, fd, ev,
						       path, buf_len);
	else if (path->ip_addr_len == 16)
		return cnic_handle_ipv6_iscsi_path_req(nic, fd, ev,
						       path, buf_len);
	else {
		LOG_DEBUG(PFX "%s: unknown ip_addr_len: %d size dropping ",
			  nic->log_name, path->ip_addr_len);
		return -EIO;
	}
}
